#!/usr/bin/env python3
import sys, os, time
import socket
import threading
import time

# list of users allowed without login
anonymous_users = ( "anonymous", "ftp" )

# override IP address for PASV commands
# set to None to disable override
pasv_ip = None
#pasv_ip = "10.0.2.2"

class FtpSession (threading.Thread):
	def __init__(self, counter, addr, socket, path):
		threading.Thread.__init__(self)
		self.id = counter
		self.root = path
		self.csocket = socket
		self.dsocket = None
		self.user = None
		self.binary = False
		self.cwd = ""
		self.offset = 0
		print("[%d] Connection from %s:%d." % (self.id, addr[0], addr[1]))
		self.send("220-Ready.")
		self.send("220 Features: p .")

	def run(self):
		data = b''
		while self.csocket:
			raw = self.csocket.recv(1024)
			if raw:
				data += raw
				while b'\n' in data:
					line = data[:data.find(b'\n')]
					data = data[len(line)+1:]
					if(line[-1] == 13):
						line = line[:-1]
					split = line.partition(b' ')
					self.handle(split[0], split[2])
			else:
				self.csocket.close()
				self.csocket = None

		print("[%d] Connection closed." % self.id)
		if self.dsocket:
			self.dsocket.close()
			self.dsocket = None

	def dconnect(self):
		if self.dsocket and self.passive:
			sock, addr = self.dsocket.accept()
			print("[%d] PASV connection from %s." % (self.id, addr[0]))
			self.dsocket.close()
			self.dsocket = sock
			self.passive = False
			return True
		elif self.dsocket and not self.passive:
			# active socket or connected
			return True
		else:
			print("[%d] Data connection missing.")
			return False

	def send(self, data):
		print("[%d] << %s" % (self.id, data))
		data += "\r\n"
		self.csocket.sendall(data.encode('iso8859-1'))

	def handle(self, cmd, args):
		print("[%d] >> %s %s" % (self.id, cmd, args))
		cmd = cmd.decode('iso8859-1').lower()

		# no login required
		# TODO: user management?
		if cmd == "quit":
			self.send("221 Bye.")
			self.csocket.close()
			self.csocket = None
		elif cmd == "user":
			self.user = args.decode('iso8859-1')
			self.send("230 OK")
		elif cmd == "pass" or cmd == "acct":
			if self.user in anonymous_users:
				self.send("202 OK")
			else:
				self.send("530 Permission denied.")
		elif cmd == "noop":
			self.send("200 OK")
		elif cmd == "syst":
			self.send("215 UNIX Type: L8")
		elif cmd == "help" and not args:
			self.send("214-Partially understood commands:")
			self.send("214-QUIT USER PASS NOOP SYST HELP")
			self.send("214-TYPE STRU MODE STAT CWD PWD CDUP")
			self.send("214-LIST NLST PORT PASV EPSV REST RETR")
			self.send("214 End of HELP.")
		elif cmd == "help":
			self.send("502 HELP not available.")
		elif not self.user in anonymous_users:
			self.send("530 Permission denied.")
			return

		# login required
		elif cmd == "type":
			if args == b'A' or args == b'A N':
				self.binary = False
				self.send("200 OK")
			elif args == b'I' or args == b'L 8':
				self.binary = True
				self.send("200 OK")
			else:
				self.send("504 Nope.")
		elif cmd == "stru":
			if args == b'F' or args == b'f':
				self.send("200 OK")
			else:
				self.send("504 Nope.")
		elif cmd == "mode":
			if args == b'S' or args == b's':
				self.send("200 OK")
			else:
				self.send("504 Nope.")
		elif cmd == "stat" and not args:
			self.send("211 User \"%s\", Path \"/%s\", Binary %s" %
				(self.user, self.cwd, self.binary))
		elif cmd == "stat":
			self.send("504 Nope.")
		elif cmd == "cwd" or cmd == "xcwd":
			arg = args.decode('iso8859-1')
			if not arg[0] == '/':
				arg = self.cwd + '/' + arg

			path = os.path.abspath(self.root + '/' + arg)
			if not path[:len(self.root)] == self.root:
				self.send("550 Nope!") # outside of root
			elif not os.path.exists(path):
				self.send("550 Nope.") # inaccessible
			elif not os.path.isdir(path):
				self.send("550 Nope?") # not a directory
			else:
				self.cwd = path[len(self.root)+1:]
				self.send("250 OK")
		elif cmd == "pwd" or cmd == "xpwd":
			self.send("257 \"/%s\"" % self.cwd)
		elif cmd == "cdup" or cmd == "xcdup":
			self.send("550 Nope.")
		elif cmd == "list" and not args:
			# get result list
			try:
				path = self.root + "/" + self.cwd
				results = []
				for name in os.listdir(path):
					isdir = os.path.isdir(path + "/" + name)
					size = os.path.getsize(path + "/" + name)
					results.append((name, isdir, size))
			except:
				self.send("451 Nope.")
				return

			# establish data connection
			if not self.dconnect():
				self.send("425 No data connection.")
				return

			# send output
			try:
				self.send("150 LIST ready.")
				for result in results:
					line  = "dr-xr-xr-x 1 owner group " if result[1] else "-r--r--r-- 1 owner group "
					line += str(result[2]).rjust(13) + " "
					line += "Jan  1 00:00 "
					line += result[0]
					line += "\r\n"
					self.dsocket.sendall(line.encode('iso8859-1'))
				resp = "226 OK"
			except:
				resp = "426 Nope."

			# close data connection
			self.dsocket.close()
			self.dsocket = None
			self.send(resp)

		elif cmd == "nlst" and not args:
			# get result list (name only)
			try:
				path = self.root + "/" + self.cwd
				results = os.listdir(path)
			except:
				self.send("451 Nope.")
				return

			# establish data connection
			if not self.dconnect():
				self.send("425 No data connection.")
				return

			# send output
			try:
				self.send("150 NLST ready.")
				for result in results:
					line = result + "\r\n"
					self.dsocket.sendall(line.encode('iso8859-1'))
				resp = "226 OK"
			except:
				resp = "426 Nope."

			# close data connection
			self.dsocket.close()
			self.dsocket = None
			self.send(resp)

		elif cmd == "list":
			self.send("504 Nope.")
		elif cmd == "nlst":
			self.send("504 Nope.")
		elif cmd == "port":
			if self.dsocket:
				self.dsocket.close()
				self.dsocket = None

			arg = args.decode('iso8859-1').split(',')
			try:
				ip = str(int(arg[0])) + "." + str(int(arg[1])) + "." + \
					str(int(arg[2])) + "." + str(int(arg[3]))
				port = 256 * int(arg[4]) + int(arg[5])
				if port < 1024: raise Exception("invalid port")
				print("[%d] PORT connection to %s:%d." % (self.id, ip, port))
				time.sleep(0.5)
				self.dsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
				self.dsocket.settimeout(10)
				self.dsocket.connect((ip, port))
				self.passive = False
				self.send("200 OK")
			except Exception as e:
				print("[%d] PORT connection failed: %s." % (self.id, e))
				self.send("425 Connection failed.")
				if self.dsocket:
					self.dsocket.close()
				self.dsocket = None
		elif cmd == "pasv" or cmd == "epsv":
			if self.dsocket:
				self.dsocket.close()
				self.dsocket = None

			myip = self.csocket.getsockname()[0]
			for port in range(40000, 41000):
				try:
					self.dsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
					self.dsocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
					self.dsocket.bind((myip, port))
					break
				except OSError as e:
					pass
			else:
				self.send("425 Connection failed.")
				self.dsocket = None
				return

			self.dsocket.listen()
			self.passive = True
			if cmd == "epsv":
				self.send("229 Entering Extended Passive Mode (|||%d|)" % port)
			else:
				ip = myip
				if pasv_ip:
					ip = pasv_ip
				self.send("227 Entering Passive Mode (%s,%i,%i)" %
					(ip.replace(".", ","), int(port/256), int(port%256)))
		elif cmd == "rest":
			try:
				self.offset = int(args.decode('iso8859-1'))
				if self.offset < 0:
					raise Exception("invalid argument")
				self.send("350 OK")
			except:
				self.offset = 0
				self.send("504 Nope.")
		elif cmd == "retr":
			# open file
			try:
				arg = args.decode('iso8859-1')
				if arg[0] == '/':
					path = self.root + arg
				else:
					path = self.root + "/" + self.cwd + "/" + arg

				if not os.path.exists(path):
					print("%s not found" % path)
					raise Exception("not found")
				fh = open(path, "rb" if self.binary else "r")
			except Exception as e:
				self.send("550 Nope.")
				return

			# establish data connection
			if not self.dconnect():
				self.send("425 No data connection.")
				return

			# send output
			try:
				self.send("150 RETR ready.")
				self.dsocket.sendall(fh.read())
				resp = "226 OK"
			except Exception as e:
				print("[%d] Failed to send: %s." % (self.id, e))
				fh.close()
				resp = "426 Nope."

			fh.close()
			self.dsocket.close()
			self.dsocket = None
			self.send(resp)
		elif cmd == "stor" or cmd == "appe" or cmd == "stou" or cmd == "allo" or \
			cmd == "mkd" or cmd == "xmkd" or cmd == "rmd" or cmd == "xrmd" or \
			cmd == "dele" or cmd == "rnfr" or cmd == "rnto":
			# no write access support
			self.send("530 Permission denied.")
		else:
			self.send("500 Nope.")

def parse_args(argv):
	server = "127.0.0.1"
	port = 21
	path = os.path.abspath(".")
	try:
		if len(argv) > 1:
			server = socket.gethostbyname(argv[1])
		if len(argv) > 2:
			port = int(argv[2])
		if len(argv) > 3:
			path = os.path.abspath(argv[3])
			if not os.path.exists(path):
				raise Exception("path does not exist")
		if len(argv) > 4:
			raise Exception("too many arguments")

	except Exception as e:
		print("Usage: %s server_ip port path" % argv[0])
		print("Error: %s" % e)
		sys.exit(1)

	return (server, port, path)

# create server socket
try:
	server, port, path = parse_args(sys.argv)
	ssocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	ssocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	ssocket.bind((server, port))
	ssocket.listen()
	counter = 0
	print("FTP ready on %s:%d, serving \"%s\"." % (server, port, path))
	while True:
		# wait for connection
		csocket, addr = ssocket.accept()

		# create session thread
		session = FtpSession(counter, addr, csocket, path)
		session.daemon = True
		counter = counter + 1
		session.start()

except KeyboardInterrupt:
	print("FTP shutdown.")
	ssocket.close()
	threads = threading.enumerate()
#	for t in threads:
#		if t.name != "MainThread" and t.is_alive():
#			print("Wait for %s to shutdown..." % t.name)
#			t.join(timeout=3.0)
#			if t.is_alive():
#				print("...still alive, will be killed.")
	if len(threads) > 1:
		print("(killing %d active sessions)" % (len(threads)-1))
	sys.exit(1)

except Exception as e:
	print("Failed to start server: %s." % e)
	sys.exit(1)
